#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "log.h"
#include "cmd56.h"
#include "f00d_emu.h"

#include "crypto/aes.h"
#include "crypto/aes_cbc.h"
#include "crypto/aes_cmac.h"

// game specific keys, used to derive the rif key,
// which is used to decrypt the game klicensee.

// example is from "Smart As."
static char RIF_KEY_BUF[0x20] = { 0x12, 0x53, 0x56, 0x29, 0x00, 0x31, 0x00, 0x3D, 0x00, 0x00, 0x00, 0x00, 0x42, 0x21, 0x00, 0x00, 0x00, 0x42, 0x02, 0x00, 0x56, 0x29, 0x05, 0x03, 0x31, 0x3D, 0x20, 0x11, 0x00, 0x00, 0x00, 0x00 };
static char KLIC_KEY_BUF[0x20] = { 0x7B, 0x2B, 0xA1, 0xF1, 0xB7, 0x57, 0xF0, 0x35, 0xFA, 0x93, 0x94, 0x0D, 0x1A, 0xB4, 0xD9, 0x1A, 0x18, 0x54, 0xD6, 0xC3, 0xCD, 0xCD, 0x5B, 0x67, 0xE1, 0x07, 0x70, 0xA4, 0x2B, 0x4F, 0xA9, 0x0A };

// work buffers, packet request and response
static char CMD56_PACKET[0x200];
static char CMD56_PACKET_RESPONSE[0x200];

// CART_RANDOM is used to derive the master_key using bbmac 0x305
// and 0x308 (only on RETAIL_KEY_ID) 
static char VITA_AUTHENTICITY_KEY[0x30];
static char CART_RANDOM[0x20];

// cart lock status, unlocked after successful VITA_AUTHENTICITY_CHECK
static uint8_t CART_STATUS = READ_WRITE_LOCK;

// KEY_ID to use, every gc ive ever seen uses RETAIL_KEY_ID (0x1).
// however a PROTOTYPE_KEY_ID1 0x8001 is also allowed,

// PROTOTYPE_KEY_ID actually has worse security in place, 
// however sony could easily check for this in a future update
// and can use RETAIL_KEY_ID thanks to the racoon exploit anyway. 

// There also exists a PROTOTYPE_KEY_ID2 and 3,0x8002 and 0x8003, 
// however these are blacklisted as of fw 1.04.
static uint16_t KEY_ID = RETAIL_KEY_ID;

#define PACKET_RESPONSE_START(resp, pack) \
	dst_packet_header* resp = (dst_packet_header*)CMD56_PACKET_RESPONSE; \
	memset(resp, 0x00, sizeof(dst_packet_header)); \
	resp->response_code = pack->response_code; \
	resp->additional_data_size = 0; \
	resp->response_size = __builtin_bswap16((unsigned short)pack->response_size); \
	resp->error_code = 0; \
	memset(response->data, 0x00, pack->response_size);

void handle_cmd_start(src_packet_header* packet){
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_cmd_start\n");

	response->data[0xD] = 0x1;
	response->data[0xE] = 0x1;
	response->data[0xF] = 0x4;
}
void handle_cmd_status(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_cmd_status\n");

	response->data[0x0] = CART_STATUS;
	response->data[0x1] = 0x0;
}

void handle_vita_authenticity_check(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_vita_authenticity_check\n");

	memcpy(VITA_AUTHENTICITY_KEY, packet->data, sizeof(VITA_AUTHENTICITY_KEY));
	LOG("Vita Authenticity Key: ");
	LOG_BUFFER(packet->data, sizeof(VITA_AUTHENTICITY_KEY));

	// just always say the vita is authentic :tm: ..
	response->data[0x0] = 0x00;
	response->data[0x1] = 0x00;
	response->data[0x3] = 0x00;

	CART_STATUS = ALL_OK;
}

void handle_generate_random_keyseed(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_generate_random_keyseed\n");

	response->data[0x0] = 0xE0;
	response->data[0x1] = 0x00;

	// specify key_id
	response->data[0x2] = (KEY_ID & 0xFF00) >> 8;
	response->data[0x3] = (KEY_ID & 0x00FF);

	response->data[0x4] = 0x00;
	response->data[0x5] = 0x02;

	response->data[0x6] = 0x00;
	response->data[0x7] = 0x03;

	// EXPLOIT:
	// dont actually use a random number, this way the key derived shall be the same every time.

	memset(CART_RANDOM, 0xAA, sizeof(CART_RANDOM));
	memcpy(response->data + 0x8, CART_RANDOM, sizeof(CART_RANDOM));
}

void handle_challenge(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_challenge\n");
	
	char secondary_key0[0x10];
	decrypt_secondary_key0(CART_RANDOM, KEY_ID, VITA_AUTHENTICITY_KEY, secondary_key0);

	LOG("secondary_key0: ");
	LOG_BUFFER(secondary_key0, sizeof(secondary_key0));

	// copy challenge data
	memcpy(response->data + 0x8, packet->data, 0x10);

	// copy cart random
	memcpy(response->data + 0x18, CART_RANDOM, sizeof(CART_RANDOM));

	LOG("Plaintext: ");
	LOG_BUFFER(response->data, 0x40);

	encrypt_with_master_key(secondary_key0, response->data, response->data, 0x40);

}
void handle_klic_part_and_cmac_signature(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_klic_part_and_cmac_signature\n");

	char secondary_key0[0x10];
	char cmac_input[0x40];

	decrypt_secondary_key0(CART_RANDOM, KEY_ID, VITA_AUTHENTICITY_KEY, secondary_key0);
	LOG("secondary_key0: ");
	LOG_BUFFER(secondary_key0, sizeof(secondary_key0));

	// decrypt challenge value
	decrypt_with_master_key(secondary_key0, response->data, packet->data, 0x20);

	// copy klic part of key
	memcpy(response->data + 0x10, KLIC_KEY_BUF, sizeof(KLIC_KEY_BUF));

	encrypt_with_master_key(secondary_key0, response->data, response->data, 0x30);

	// aes-128-cmac the whole thing
	memcpy(cmac_input, &response->response_size, 0x3);
	memset(cmac_input + 0x3, 0x00, 0xD);
	memcpy(cmac_input + 0x10, response->data, 0x30);

	aes_cmac(cmac_input, sizeof(cmac_input), secondary_key0, response->data + 0x30);

}


void handle_rif_buf_part_hash_key(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);

	LOG("handle_rif_buf_part_hash_key\n");

	char secondary_key0[0x10];
	char cmac_input[0x50];

	decrypt_secondary_key0(CART_RANDOM, KEY_ID, VITA_AUTHENTICITY_KEY, secondary_key0);
	LOG("secondary_key0: ");
	LOG_BUFFER(secondary_key0, sizeof(secondary_key0));
	
	// copy random value
	memcpy(response->data+0x8, packet->data, 0x10);

	// copy klic buffer
	memcpy(response->data + 0x18, RIF_KEY_BUF, sizeof(RIF_KEY_BUF));

	LOG("plaintext response: ");
	LOG_BUFFER(response->data, 0x40);

	// encrypt the response
	encrypt_with_master_key(secondary_key0, response->data, response->data, 0x40);

	// aes-128-cmac the whole thing
	memcpy(cmac_input, &response->response_size, 0x3);
	memset(cmac_input + 0x3, 0x00, 0xD);
	memcpy(cmac_input + 0x10, response->data, 0x40);

	aes_cmac(cmac_input, sizeof(cmac_input), secondary_key0, response->data + 0x40);
	LOG("CMAC: ");
	LOG_BUFFER(response->data + 0x40, 0x10);
}

void handle_vita_random(src_packet_header* packet) {
	PACKET_RESPONSE_START(response, packet);
	LOG("handle_vita_random\n");

	char vita_random[0x10];
	memcpy(vita_random, packet->data + 0x2, sizeof(vita_random));
	
	LOG("Vita Random: ");
	LOG_BUFFER(vita_random, sizeof(vita_random));

	char master_key[0x10];
	derive_master_key(master_key, CART_RANDOM, KEY_ID);

	LOG("Master Key: ");
	LOG_BUFFER(master_key, sizeof(master_key));

	memcpy(response->data + 0x10, vita_random, sizeof(vita_random));

	LOG("Plaintext: ");
	LOG_BUFFER(response->data, 0x20);

	encrypt_with_master_key(master_key, response->data, response->data, 0x20);

}


void handle_packet(src_packet_header* packet) {
	if(memcmp(packet->key, key0, sizeof(packet->key)) == 0) {
		switch(packet->command) {
			case START: //packet1, packet2
				handle_cmd_start(packet); 
				break;
			case GET_STATUS: // packet3, packet4
				handle_cmd_status(packet); 
				break;
			case GENERATE_RANDOM_KEYSEED: // packet5, packet6
				handle_generate_random_keyseed(packet); 
				break;
			case ENCRYPT_VITA_RANDOM: // packet7, packet8
				handle_vita_random(packet); 
				break;
			case VITA_AUTHENTICITY_CHECK: // packet9, packet10
				handle_vita_authenticity_check(packet); 
				break;
				
			// packet11, packet12 -> GET_STATUS again
			
			case CHALLENGE: // packet13, packet14
				handle_challenge(packet);
				break;

			// packet15, packet16 -> CHALLENGE again

			case KLIC_PART_AND_CMAC_SIGNATURE: // packet17, packet18
				handle_klic_part_and_cmac_signature(packet); 
				break;
			case RIFBUF_KEY_PART: // packet19, packet20
				handle_rif_buf_part_hash_key(packet); 
				break;
			default:
				LOG("Unknown command: %x\n", packet->command);
				break;
		}
	}
}

// exposed functions :

void cmd56_reset() {
	// lock cart for reading/writing
	CART_STATUS = READ_WRITE_LOCK;

	// set default key id
	KEY_ID = RETAIL_KEY_ID;
	
	// reset keys derived from CMD56 algorithm
	memset(VITA_AUTHENTICITY_KEY, 0x00, sizeof(VITA_AUTHENTICITY_KEY));
	memset(CART_RANDOM, 0x00, sizeof(CART_RANDOM));
	
	// reset packet data
	memset(CMD56_PACKET, 0x00, sizeof(CMD56_PACKET));
	memset(CMD56_PACKET_RESPONSE, 0x00, sizeof(CMD56_PACKET_RESPONSE));
	
	// reset cart specific keys
	memset(RIF_KEY_BUF, 0x00, sizeof(RIF_KEY_BUF));
	memset(KLIC_KEY_BUF, 0x00, sizeof(KLIC_KEY_BUF));
}

void cmd56_update_keyid(uint16_t key_id) {
	KEY_ID = key_id;
}

void cmd56_set_keys(char* rif_part, char* klic_part) {
	memcpy(RIF_KEY_BUF, rif_part, sizeof(RIF_KEY_BUF));
	memcpy(KLIC_KEY_BUF, klic_part, sizeof(KLIC_KEY_BUF));
}

void cmd56_run(char* buffer) {
	memcpy(CMD56_PACKET, buffer, sizeof(CMD56_PACKET));
	handle_packet((src_packet_header*) CMD56_PACKET);
	memcpy(buffer, CMD56_PACKET_RESPONSE, sizeof(CMD56_PACKET_RESPONSE));
}